/*
 * Touchscreen Key Driver
 * Allows areas of a touchscreen to be used as a keyboard
 * passing the keypresses to the system via a uinput device.
 *
 * Copyright (c) 2005 Alberto Mardegan 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include "tslib.h"

#define MAX_OFFSCREEN_KEYS  20

struct rectangle {
	int16_t x1;
	int16_t y1;
	int16_t x2;
	int16_t y2;
};

struct os_keys {
	struct rectangle bb;           /* bounding box of all keys */
	int n_keys;
	struct os_key {
		__u16 code;
		struct rectangle r;
	} key[MAX_OFFSCREEN_KEYS];
};
 

struct uinput_user_dev uud = {
	.name = "Touchscreen offscreen keys",
	.id = { BUS_XTKBD, 1, 1, 1 },
};


int read_config_file(char *filename, struct os_keys *keys)
{
	FILE *f;
	int n_conv;
	int16_t h, w;
	struct os_key *k;
	char line[100];

	f = fopen(filename, "r");
	if (!f) {
		fprintf(stderr, "Error opening %s: %s\n", filename, strerror(errno));
		return -1;
	}

	keys->bb.x1 = keys->bb.y1 = 0x7fff;
	keys->bb.x2 = keys->bb.y2 = 0;
	keys->n_keys = 0;
	k = keys->key;
	while (fgets(line, sizeof(line), f) != 0) {
		if (line[0] == '#') 
			continue;
		if (keys->n_keys == MAX_OFFSCREEN_KEYS) 
			break;
		n_conv = sscanf(line, "%hd,%hd,%hd,%hd,%hd",
			&k->code, &k->r.x1, &k->r.y1, &w, &h);
		if (n_conv != 5) {
			fprintf(stderr, "Error parsing line: %s", line);
			return -2;
		}
		k->r.x2 = k->r.x1 + w;
		k->r.y2 = k->r.y1 + h;
		/* update the bounding box */
		if (k->r.x1 < keys->bb.x1) 
			keys->bb.x1 = k->r.x1;
		if (k->r.y1 < keys->bb.y1)
			keys->bb.y1 = k->r.y1;
		if (k->r.x2 > keys->bb.x2)
			keys->bb.x2 = k->r.x2;
		if (k->r.y2 > keys->bb.y2)
			keys->bb.y2 = k->r.y2;
		keys->n_keys++;
		k++;
	}
	fclose(f);
	return keys->n_keys;
}


int key_from_xy(struct os_keys *keys, struct ts_sample *samp)
{
	struct os_key *k;
	int i_key;

	/* check it the point lies inside the bounding box */
	if (samp->x < keys->bb.x1 || samp->x > keys->bb.x2 ||
	    samp->y < keys->bb.y1 || samp->y > keys->bb.y2)
		return 0;
	for (i_key = 0, k = keys->key; i_key < keys->n_keys; i_key++, k++)
		if (samp->x >= k->r.x1 && samp->x <= k->r.x2 &&
		    samp->y >= k->r.y1 && samp->y <= k->r.y2)
			return k->code;
	return 0;
}


int open_uinput()
{
	char *uinput_names[] = { "/dev/uinput", "/dev/input/uinput", 0 };
	char **uinput_name;
	int fd;

	for (uinput_name = uinput_names; *uinput_name != NULL; uinput_name++) {
		if (access(*uinput_name, W_OK) == 0) 
			break;
		if (*uinput_name == NULL) {
			fputs("No valid uinput devices found; is uinput module loaded?\n", stderr);
			return -1;
		}
	}

	fd = open(*uinput_name, O_RDWR);
	if (!fd) {
		fprintf(stderr, "Error opening %s: %s\n", *uinput_name, strerror(errno));
		return -1;
	}
	return fd;
}


void press_key(int fd, int key, int pressed)
{
	struct input_event ev;
	int r;

	ev.time.tv_sec = time(0);
	ev.time.tv_usec = 0;
	ev.type = EV_KEY;
	ev.code = key;
	ev.value = pressed;
	r = write(fd, &ev, sizeof(ev));
	if (r != sizeof(ev))
		perror("Error writing to uinput");
}


int main(int argc, char**argv)
{
	struct tsdev *ts;
	struct ts_sample samp;
	struct os_keys keys;
	struct os_key *k;
	int fd, r, i_key;
	int key, last_key = 0;

	if (argc != 3) {
		fprintf(stderr, "Usage: %s map_file ts_device\n", argv[0]);
		return 1;
	}

	if (read_config_file(argv[1], &keys) <= 0) 
		return 1;

	ts = ts_open(argv[2], 0);
	if (!ts) {
		fprintf(stderr, "Error opening %s: %s\n", argv[2], strerror(errno));
		return 1;
	}
	ts_config(ts);

	fd = open_uinput();
	if (fd < 0) 
		return 1;

	r = write(fd, &uud, sizeof(uud));
	if (r == -1) {
		perror("Error initializing uinput device");
		return 1;
	}

	ioctl(fd, UI_SET_EVBIT, EV_KEY);

	for (i_key = 0, k = keys.key; i_key < keys.n_keys; i_key++, k++)
		ioctl(fd, UI_SET_KEYBIT, k->code);

	r = ioctl(fd, UI_DEV_CREATE);
	if (r == -1) {
		perror("Error creating uinput device");
		return 1;
	}

	while (ts_read(ts, &samp, 1) >= 0) {
		key = key_from_xy(&keys, &samp);

		/* if the stylus is held down and moved (to another button or out of
		 * the bounding box) we want to release the key that was pressed last
		 */
    		if (key != last_key && last_key != 0 && samp.pressure > 0)
	    		press_key(fd, last_key, 0);

    		if (key != 0)
			press_key(fd, key, (samp.pressure > 0) ? 1 : 0);
		last_key = key;
	}
	ts_close(ts);
	r = ioctl(fd, UI_DEV_DESTROY);
	if (r == -1)
		perror("Error destroying uinput device");
	close(fd);

	return 0;
}

